#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#ifdef USE_MATHLIB
#include <math.h>
#endif

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "config.h"
#include "types.h"
#include "lib.h"
#include "value.h"
#include "area.h"
#include "function.h"
#include "framebuf.h"
#include "term.h"

static int function_puts(value_t value, value_t args);
static int function_input(value_t value, value_t args);
static int function_screen(value_t value, value_t args);
static int function_scrclr(value_t value, value_t args);
static int function_getscr(value_t value, value_t args);
static int function_getpos(value_t value, value_t args);
static int function_attribute(value_t value, value_t args);
static int function_move(value_t value, value_t args);
static int function_scroll(value_t value, value_t args);
static int function_getch(value_t value, value_t args);
static int function_inkey(value_t value, value_t args);
static int function_itoa(value_t value, value_t args);
static int function_atoi(value_t value, value_t args);
#ifdef NLL_FLOATING_POINT
static int function_itof(value_t value, value_t args);
static int function_ftoi(value_t value, value_t args);
#endif
static int function_length(value_t value, value_t args);
static int function_left(value_t value, value_t args);
static int function_right(value_t value, value_t args);
static int function_mid(value_t value, value_t args);
static int function_instr(value_t value, value_t args);
static int function_strcpy(value_t value, value_t args);
static int function_rand(value_t value, value_t args);
static int function_ascii(value_t value, value_t args);
static int function_char(value_t value, value_t args);
static int function_int(value_t value, value_t args);
static int function_abs(value_t value, value_t args);
static int function_sign(value_t value, value_t args);
static int function_time(value_t value, value_t args);
static int function_date(value_t value, value_t args);
static int function_memory(value_t value, value_t args);
static int function_peek(value_t value, value_t args);
static int function_poke(value_t value, value_t args);
static int function_read(value_t value, value_t args);
static int function_write(value_t value, value_t args);
static int function_memcpy(value_t value, value_t args);
#ifdef USE_MATHLIB
static int function_pi(value_t value, value_t args);
static int function_sin(value_t value, value_t args);
static int function_cos(value_t value, value_t args);
static int function_sqr(value_t value, value_t args);
#endif
static int function_fopen(value_t value, value_t args);
static int function_fclose(value_t value, value_t args);
static int function_fgets(value_t value, value_t args);
static int function_fputs(value_t value, value_t args);
static int function_fread(value_t value, value_t args);
static int function_fwrite(value_t value, value_t args);

struct function functions[] = {
  { "PUTS"     , NULL  , FUNCTION_TYPE_PUTS     , 1, function_puts      },
  { "INPUT"    , NULL  , FUNCTION_TYPE_INPUT    , 1, function_input     },
  { "SCREEN"   , "SCR" , FUNCTION_TYPE_SCREEN   , 1, function_screen    },
  { "SCRCLR"   , NULL  , FUNCTION_TYPE_SCRCLR   , 1, function_scrclr    },
  { "GETSCR"   , NULL  , FUNCTION_TYPE_GETSCR   , 3, function_getscr    },
  { "GETPOS"   , NULL  , FUNCTION_TYPE_GETPOS   , 2, function_getpos    },
  { "ATTRIBUTE", "ATTR", FUNCTION_TYPE_ATTRIBUTE, 1, function_attribute },
  { "MOVE"     , NULL  , FUNCTION_TYPE_MOVE     , 1, function_move      },
  { "SCROLL"   , NULL  , FUNCTION_TYPE_SCROLL   , 1, function_scroll    },
  { "GETCH"    , NULL  , FUNCTION_TYPE_GETCH    , 1, function_getch     },
  { "INKEY"    , "KEY" , FUNCTION_TYPE_INKEY    , 1, function_inkey     },
  { "ITOA"     , "STR" , FUNCTION_TYPE_ITOA     , 1, function_itoa      },
  { "ATOI"     , "VAL" , FUNCTION_TYPE_ATOI     , 1, function_atoi      },
#ifdef NLL_FLOATING_POINT
  { "ITOF"     , NULL  , FUNCTION_TYPE_ITOF     , 1, function_itof      },
  { "FTOI"     , NULL  , FUNCTION_TYPE_FTOI     , 1, function_ftoi      },
#endif
  { "LENGTH"   , "LEN" , FUNCTION_TYPE_LENGTH   , 1, function_length    },
  { "LEFT"     , NULL  , FUNCTION_TYPE_LEFT     , 1, function_left      },
  { "RIGHT"    , NULL  , FUNCTION_TYPE_RIGHT    , 1, function_right     },
  { "MID"      , NULL  , FUNCTION_TYPE_MID      , 1, function_mid       },
  { "INSTR"    , NULL  , FUNCTION_TYPE_INSTR    , 1, function_instr     },
  { "STRCPY"   , NULL  , FUNCTION_TYPE_STRCPY   , 1, function_strcpy    },
  { "RAND"     , "RND" , FUNCTION_TYPE_RAND     , 1, function_rand      },
  { "ASCII"    , "ASC" , FUNCTION_TYPE_ASCII    , 1, function_ascii     },
  { "CHAR"     , "CHR" , FUNCTION_TYPE_CHAR     , 1, function_char      },
  { "INT"      , NULL  , FUNCTION_TYPE_INT      , 1, function_int       },
  { "ABS"      , NULL  , FUNCTION_TYPE_ABS      , 1, function_abs       },
  { "SIGN"     , "SGN" , FUNCTION_TYPE_SIGN     , 1, function_sign      },
  { "TIME"     , NULL  , FUNCTION_TYPE_TIME     , 1, function_time      },
  { "DATE"     , NULL  , FUNCTION_TYPE_DATE     , 1, function_date      },
  { "MEMORY"   , NULL  , FUNCTION_TYPE_MEMORY   , 1, function_memory    },
  { "PEEK"     , NULL  , FUNCTION_TYPE_PEEK     , 1, function_peek      },
  { "POKE"     , NULL  , FUNCTION_TYPE_POKE     , 1, function_poke      },
  { "READ"     , NULL  , FUNCTION_TYPE_READ     , 1, function_read      },
  { "WRITE"    , NULL  , FUNCTION_TYPE_WRITE    , 1, function_write     },
  { "MEMCPY"   , NULL  , FUNCTION_TYPE_MEMCPY   , 1, function_memcpy    },
#ifdef USE_MATHLIB
  { "PI"       , NULL  , FUNCTION_TYPE_PI       , 1, function_pi        },
  { "SIN"      , NULL  , FUNCTION_TYPE_SIN      , 1, function_sin       },
  { "COS"      , NULL  , FUNCTION_TYPE_COS      , 1, function_cos       },
  { "SQR"      , NULL  , FUNCTION_TYPE_SQR      , 1, function_sqr       },
#endif
  { "FOPEN"    , NULL  , FUNCTION_TYPE_FOPEN    , 1, function_fopen     },
  { "FCLOSE"   , NULL  , FUNCTION_TYPE_FCLOSE   , 1, function_fclose    },
  { "FGETS"    , NULL  , FUNCTION_TYPE_FGETS    , 1, function_fgets     },
  { "FPUTS"    , NULL  , FUNCTION_TYPE_FPUTS    , 1, function_fputs     },
  { "FREAD"    , NULL  , FUNCTION_TYPE_FREAD    , 1, function_fread     },
  { "FWRITE"   , NULL  , FUNCTION_TYPE_FWRITE   , 1, function_fwrite    },
  { NULL       , NULL  , FUNCTION_TYPE_NONE     , 0, NULL               }
};

const char *function_name(function_t function)
{
  return function->name;
}

/* #define NLL_FUNCTION_SHORTNAME */

int function_search(const char *name, function_t *functionp)
{
  function_t function;
#ifdef NLL_FUNCTION_SHORTNAME
  int len;
#endif

  for (function = functions; function->name; function++) {
    if (!strcmp(function->name, name)) {
      *functionp = function;
      return 0;
    }
  }

  for (function = functions; function->name; function++) {
    if (function->other_name && !strcmp(function->other_name, name)) {
      *functionp = function;
      return 0;
    }
  }

#ifdef NLL_FUNCTION_SHORTNAME
  for (len = 0; name[len]; len++) {
    if ((name[len] == '.') && !name[len + 1]) {
      for (function = functions; function->name; function++) {
	if (!strncmp(function->name, name, len)) {
	  *functionp = function;
	  return 0;
	}
      }
      break;
    }
  }
#endif

  return NLL_ERRCODE_FORMULA_UNKNOWN_FUNCTION;
}

int function_proc(function_t function, value_t value, value_t args)
{
  return function->function(value, args);
}

static int function_puts(value_t value, value_t args)
{
  int r, n = 0;

  value_print(args);

  for (; args; args = args->next, n++)
    ;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, n)) < 0)
    return r;

  return 0;
}

static int function_input(value_t value, value_t args)
{
  int r, mode, len;
  char *string;
  char buffer[STRING_MAXLEN + 1];

  mode = framebuf_get_screen_mode();
  if (mode)
    framebuf_setscreen(0, -1, -1, -1, -1);

  for (; args; args = args->next) {
    if (value_get_string(args, &string, NULL) < 0)
      continue;
    printf("%s", string);
  }
  fflush(stdout);

  if (mode)
    framebuf_setscreen(mode, -1, -1, -1, -1);

  nll_term_restore();

  if (!fgets(buffer, sizeof(buffer), stdin))
    buffer[0] = '\0';

  len = strlen(buffer);
  while ((len > 0) && (buffer[len - 1] == '\n'))
    buffer[--len] = '\0';

  nll_term_start();

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, buffer, -1)) < 0)
    return r;

  return 0;
}

static int function_screen(value_t value, value_t args)
{
  int r, n = 0;
  integer_t integer;
  int mode = -1, width = -1, height = -1, flush = -1, cursol = -1;

  for (; args; args = args->next, n++) {
    if (value_get_integer(args, &integer) < 0)
      continue;
    switch (n) {
    case 0: mode   = integer; break;
    case 1: width  = integer; break;
    case 2: height = integer; break;
    case 3: flush  = integer; break;
    case 4: cursol = integer; break;
    default:
      return NLL_ERRCODE_FORMULA_MUCH_PARAMETER;
    }
  }

  framebuf_setscreen(mode, width, height, flush, cursol);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, n)) < 0)
    return r;

  return 0;
}

static int function_scrclr(value_t value, value_t args)
{
  int r, n = 0;
  integer_t integer;
  int ch = -1, flush = -1;

  for (; args; args = args->next, n++) {
    if (value_get_integer(args, &integer) < 0)
      continue;
    switch (n) {
    case 0: ch    = integer; break;
    case 1: flush = integer; break;
    default:
      return NLL_ERRCODE_FORMULA_MUCH_PARAMETER;
    }
  }

  framebuf_clear(ch, flush);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, n)) < 0)
    return r;

  return 0;
}

static int function_getscr(value_t value, value_t args)
{
  int r;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, framebuf_get_screen_mode())) < 0)
    return r;
  value = value->next;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, framebuf_get_screen_width())) < 0)
    return r;
  value = value->next;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, framebuf_get_screen_height())) < 0)
    return r;

  return 0;
}

static int function_getpos(value_t value, value_t args)
{
  int r;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, framebuf_get_cursol_x())) < 0)
    return r;
  value = value->next;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, framebuf_get_cursol_y())) < 0)
    return r;

  return 0;
}

static int function_attribute(value_t value, value_t args)
{
  int r, n = 0;
  integer_t integer;

  for (; args; args = args->next, n++) {
    if (value_get_integer(args, &integer) < 0)
      continue;
    switch (n) {
    case 0: framebuf_reset(); break;
    case 1: framebuf_setcolor(integer); break;
    case 2: framebuf_setattr(integer); break;
    default:
      return NLL_ERRCODE_FORMULA_MUCH_PARAMETER;
    }
  }

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, n)) < 0)
    return r;

  return 0;
}

static int function_move(value_t value, value_t args)
{
  int r, n = 0;
  integer_t integer;
  int x = -1, y = -1;

  for (; args && (n < 2); args = args->next, n++) {
    if (value_get_integer(args, &integer) < 0)
      continue;
    switch (n) {
    case 0: x = integer; break;
    case 1: y = integer; break;
    default:
      break;
    }
  }

  if ((x >= 0) || (y >= 0)) framebuf_move_cursol(x, y);

  for (; args; args = args->next, n++) {
    if (value_get_integer(args, &integer) < 0)
      continue;
    switch (n) {
    case 2:
      if (integer < 0) framebuf_move_left(-integer);
      else             framebuf_move_right(integer);
      break;
    case 3:
      if (integer < 0) framebuf_move_up(-integer);
      else             framebuf_move_down(integer);
      break;
    default:
      return NLL_ERRCODE_FORMULA_MUCH_PARAMETER;
    }
  }

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, n)) < 0)
    return r;

  return 0;
}

static int function_scroll(value_t value, value_t args)
{
  int r, n = 0;
  integer_t integer;

  for (; args; args = args->next, n++) {
    if (value_get_integer(args, &integer) < 0)
      continue;
    switch (n) {
    case 0:
      if (integer < 0) framebuf_scroll_up(-integer);
      else             framebuf_scroll_down(integer);
      break;
    default:
      return NLL_ERRCODE_FORMULA_MUCH_PARAMETER;
    }
  }

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, n)) < 0)
    return r;

  return 0;
}

static int function_getch(value_t value, value_t args)
{
  int r, n = 0;
  integer_t integer;
  int x = -1, y = -1, ch;
  char s[2];

  for (; args; args = args->next, n++) {
    if (value_get_integer(args, &integer) < 0)
      continue;
    switch (n) {
    case 0: x = integer; break;
    case 1: y = integer; break;
    default:
      return NLL_ERRCODE_FORMULA_MUCH_PARAMETER;
    }
  }

  ch = framebuf_getch(x, y);

  s[0] = ch;
  s[1] = '\0';

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, s, -1)) < 0)
    return r;

  return 0;
}

static int function_inkey(value_t value, value_t args)
{
  int r;
  integer_t integer;
  char key[2], *p;

  integer = nll_term_key();

  switch (integer) {
  case NLL_TERM_KEY_END:      p = "END"     ; break;
  case NLL_TERM_KEY_UP:       p = "UP"      ; break;
  case NLL_TERM_KEY_DOWN:     p = "DOWN"    ; break;
  case NLL_TERM_KEY_RIGHT:    p = "RIGHT"   ; break;
  case NLL_TERM_KEY_LEFT:     p = "LEFT"    ; break;
  case NLL_TERM_KEY_PAGEUP:   p = "PAGEUP"  ; break;
  case NLL_TERM_KEY_PAGEDOWN: p = "PAGEDOWN"; break;
  default:
    key[0] = integer;
    key[1] = '\0';
    p = key;
  }

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, p, -1)) < 0)
    return r;

  return 0;
}

static int function_itoa(value_t value, value_t args)
{
  int r;
  integer_t integer = 0;
  char buffer[STRING_MAXLEN + 1];
  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;
  sprintf(buffer, "%ld", integer);
  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, buffer, -1)) < 0)
    return r;
  return 0;
}

static int function_atoi(value_t value, value_t args)
{
  int r;
  integer_t integer;
  char *string;
  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &string, NULL)) < 0)
    return r;
  integer = (int)strtol(string, NULL, 0);
  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, integer)) < 0)
    return r;
  return 0;
}

#ifdef NLL_FLOATING_POINT
static int function_itof(value_t value, value_t args)
{
  int r;
  integer_t integer;
  double f;
  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;
  f = (double)integer;
  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_float(value, f)) < 0)
    return r;
  return 0;
}

static int function_ftoi(value_t value, value_t args)
{
  int r;
  integer_t integer;
  double f;
  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_float(args, &f)) < 0)
    return r;
  integer = (int)f;
  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, integer)) < 0)
    return r;
  return 0;
}
#endif

static int function_length(value_t value, value_t args)
{
  int r, length;
  char *string;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  if ((r = value_get_string(args, &string, &length)) < 0)
    return r;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  if ((r = value_set_integer(value, length)) < 0)
    return r;

  return 0;
}

static int function_left(value_t value, value_t args)
{
  int r, i, length;
  integer_t integer = 1;
  char *s, *p;
  char buffer[STRING_MAXLEN + 1];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &s, &length)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &integer)) < 0)
      return r;
  }

  p = buffer;
  for (i = 0; (i < length) && (i < integer); i++) {
    *(p++) = *(s++);
  }
  *p = '\0';

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, buffer, i)) < 0)
    return r;

  return 0;
}

static int function_right(value_t value, value_t args)
{
  int r, i, length;
  integer_t integer = 1;
  char *s, *p;
  char buffer[STRING_MAXLEN + 1];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &s, &length)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &integer)) < 0)
      return r;
  }

  s += length - integer;
  p = buffer;
  for (i = 0; i < integer; i++) {
    *(p++) = *(s++);
  }
  *p = '\0';

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, buffer, i)) < 0)
    return r;

  return 0;
}

static int function_mid(value_t value, value_t args)
{
  int r, i, length;
  integer_t integer0 = 1, integer1 = -1;
  char *s, *p;
  char buffer[STRING_MAXLEN + 1];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &s, &length)) < 0)
    return r;
  args = args->next;

  if (args) {
    if ((r = value_get_integer(args, &integer0)) < 0)
      return r;
    args = args->next;
  }

  if (args) {
    if ((r = value_get_integer(args, &integer1)) < 0)
      return r;
  }

  s += integer0;
  p = buffer;
  for (i = 0; (integer0 + i < length) && ((integer1 < 0) || (i < integer1)); i++) {
    *(p++) = *(s++);
  }
  *p = '\0';

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, buffer, i)) < 0)
    return r;

  return 0;
}

static int function_instr(value_t value, value_t args)
{
  int r, i, n, length0, length1;
  char *s0, *s1;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &s0, &length0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &s1, &length1)) < 0)
    return r;

  n = -1;
  for (i = 0; i <= length0 - length1; i++) {
    if (!memcmp(s0 + i, s1, length1)) {
      n = i;
      break;
    }
  }

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, n)) < 0)
    return r;

  return 0;
}

static int function_strcpy(value_t value, value_t args)
{
  int r, length0, length1;
  integer_t offset = 0;
  char *s0, *s1;
  char buffer[STRING_MAXLEN + 1];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &s0, &length0)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &offset)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &s1, &length1)) < 0)
    return r;

  if (offset < 0)
    offset = length0;

  if (offset + length1 > STRING_MAXLEN)
    return NLL_ERRCODE_STRING_TOO_LONG;

  memcpy(buffer, s0, length0);
  memcpy(buffer + offset, s1, length1);

  if (offset + length1 > length0)
    length0 = offset + length1;

  buffer[length0] = '\0';

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, buffer, length0)) < 0)
    return r;

  return 0;
}

static int function_rand(value_t value, value_t args)
{
  int r, n = 0;
  integer_t integer;
  static unsigned int seed = 1;

  if (args) {
    if ((r = value_get_integer(args, &integer)) < 0)
      return r;
    n = integer;
    args = args->next;
  }

  if (args) {
    if (!value_get_integer(args, &integer))
      seed = !integer ? time(NULL) : integer;
  }

  seed ^= seed << 13;
  seed ^= seed >> 17;
  seed ^= seed <<  5;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, n ? (seed % n) : seed)) < 0)
    return r;

  return 0;
}

static int function_ascii(value_t value, value_t args)
{
  int r;
  integer_t integer;
  char *string;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &string, NULL)) < 0)
    return r;

  integer = string[0];

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, integer)) < 0)
    return r;

  return 0;
}

static int function_char(value_t value, value_t args)
{
  int r;
  integer_t integer;
  char str[2];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;

  str[0] = integer;
  str[1] = '\0';

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, str, -1)) < 0)
    return r;

  return 0;
}

static int function_int(value_t value, value_t args)
{
  int r;
  integer_t integer;
#ifdef NLL_FLOATING_POINT
  double f;
#endif

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0) {
#ifdef NLL_FLOATING_POINT
    if ((r = value_get_float(args, &f)) < 0)
      return r;
    integer = f;
    while (integer > f)
      integer--;
#else
    return r;
#endif
  }

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, integer)) < 0)
    return r;

  return 0;
}

static int function_abs(value_t value, value_t args)
{
  int r;
  integer_t integer;
#ifdef NLL_FLOATING_POINT
  double f;
#endif

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0) {
#ifdef NLL_FLOATING_POINT
    if ((r = value_get_float(args, &f)) < 0)
      return r;
    if ((r = value_set_float(value, (f < 0.0) ? -f : f)) < 0)
      return r;
    return 0;
#else
    return r;
#endif
  }

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, (integer < 0) ? -integer : integer)) < 0)
    return r;

  return 0;
}

static int function_sign(value_t value, value_t args)
{
  int r;
  integer_t integer;
#ifdef NLL_FLOATING_POINT
  double f;
#endif

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0) {
#ifdef NLL_FLOATING_POINT
    if ((r = value_get_float(args, &f)) < 0)
      return r;
    if      (f < 0.0) f = -1.0;
    else if (f > 0.0) f =  1.0;
    else              f =  0.0;
    if ((r = value_set_float(value, f)) < 0)
      return r;
    return 0;
#else
    return r;
#endif
  }

  if      (integer < 0) integer = -1;
  else if (integer > 0) integer =  1;
  else                  integer =  0;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, integer)) < 0)
    return r;

  return 0;
}

static int function_time(value_t value, value_t args)
{
  int r, n;
  time_t t;

  t = time(NULL);
  n = t;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, n)) < 0)
    return r;

  return 0;
}

static int function_date(value_t value, value_t args)
{
  int r;
#ifndef USE_NLLIBC
  time_t t;
#endif
  char *s;

#ifndef USE_NLLIBC
  t = time(NULL);
  s = ctime(&t);
#else
  s = "Cannot get date.";
#endif

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, s, -1)) < 0)
    return r;

  return 0;
}

static int function_memory(value_t value, value_t args)
{
  int r;
  integer_t integer;
  area_t area;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;

  if ((r = area_alloc(&area, integer)) < 0)
    return r;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_area(value, area)) < 0)
    return r;

  return 0;
}

static int function_peek(value_t value, value_t args)
{
  int r;
  integer_t offset, size, integer;
  area_t area;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &area)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &offset)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &size)) < 0)
    return r;

  if ((r = area_read_integer(area, offset, &integer, size)) < 0)
    return r;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, integer)) < 0)
    return r;

  return 0;
}

static int function_poke(value_t value, value_t args)
{
  int r;
  integer_t offset, size, integer;
  area_t area;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &area)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &offset)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &size)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;

  if ((r = area_write_integer(area, offset, integer, size)) < 0)
    return r;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, integer)) < 0)
    return r;

  return 0;
}

static int function_read(value_t value, value_t args)
{
  int r;
  integer_t offset, size;
  area_t area;
  char buffer[STRING_MAXLEN + 1];

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &area)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &offset)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &size)) < 0)
    return r;

  if (size > STRING_MAXLEN)
    return NLL_ERRCODE_STRING_TOO_LONG;

  if ((r = area_read(area, offset, buffer, size)) < 0)
    return r;
  buffer[size] = '\0';

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, buffer, size)) < 0)
    return r;

  return 0;
}

static int function_write(value_t value, value_t args)
{
  int r, length;
  integer_t offset, size;
  area_t area;
  char *string;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &area)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &offset)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &size)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &string, &length)) < 0)
    return r;

  if (size < 0)
    size = length;

  if (length < size)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  if ((r = area_write(area, offset, string, size)) < 0)
    return r;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, size)) < 0)
    return r;

  return 0;
}

static int function_memcpy(value_t value, value_t args)
{
  int r;
  integer_t doffset, soffset, size;
  area_t darea, sarea;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &darea)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &doffset)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &sarea)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &soffset)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &size)) < 0)
    return r;

  if (size < 0)
    size = area_get_size(sarea) - soffset;

  if ((r = area_copy_area(darea, doffset, sarea, soffset, size)) < 0)
    return r;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, size)) < 0)
    return r;

  return 0;
}

#ifdef USE_MATHLIB
static int function_pi(value_t value, value_t args)
{
#ifndef NLL_FLOATING_POINT
  return NLL_ERRCODE_FORMULA_INVALID_PARAMETER;
#else
  int r;
  integer_t integer;
  double f = 1.0;
  if (args) {
    if ((r = value_get_float(args, &f)) < 0) {
      if ((r = value_get_integer(args, &integer)) < 0)
	return r;
      f = integer;
    }
  }
  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_float(value, M_PI * f)) < 0)
    return r;
  return 0;
#endif
}

static int function_sin(value_t value, value_t args)
{
#ifndef NLL_FLOATING_POINT
  return NLL_ERRCODE_FORMULA_INVALID_PARAMETER;
#else
  int r;
  integer_t integer;
  double f;
  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_float(args, &f)) < 0) {
    if ((r = value_get_integer(args, &integer)) < 0)
      return r;
    f = integer;
  }
  f = sin(f);
  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_float(value, f)) < 0)
    return r;
  return 0;
#endif
}

static int function_cos(value_t value, value_t args)
{
#ifndef NLL_FLOATING_POINT
  return NLL_ERRCODE_FORMULA_INVALID_PARAMETER;
#else
  int r;
  integer_t integer;
  double f;
  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_float(args, &f)) < 0) {
    if ((r = value_get_integer(args, &integer)) < 0)
      return r;
    f = integer;
  }
  f = cos(f);
  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_float(value, f)) < 0)
    return r;
  return 0;
#endif
}

static int function_sqr(value_t value, value_t args)
{
#ifndef NLL_FLOATING_POINT
  return NLL_ERRCODE_FORMULA_INVALID_PARAMETER;
#else
  int r;
  integer_t integer;
  double f;
  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_float(args, &f)) < 0) {
    if ((r = value_get_integer(args, &integer)) < 0)
      return r;
    f = integer;
  }
  f = sqrt(f);
  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_float(value, f)) < 0)
    return r;
  return 0;
#endif
}
#endif

static int function_fopen(value_t value, value_t args)
{
  int r;
  char *filename, *mode = "r";
  FILE *fp;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &filename, NULL)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &mode, NULL)) < 0)
    return r;

  fp = fopen(filename, mode);
  if (!fp)
    return NLL_ERRCODE_FORMULA_INVALID_PARAMETER;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, (integer_t)fp)) < 0)
    return r;

  return 0;
}

static int function_fclose(value_t value, value_t args)
{
  int r;
  integer_t integer;
  FILE *fp;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;

  fp = (FILE *)integer;
  if (fclose(fp))
    return NLL_ERRCODE_FORMULA_INVALID_PARAMETER;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, 0)) < 0)
    return r;

  return 0;
}

static int function_fgets(value_t value, value_t args)
{
  int r;
  integer_t integer;
  char buffer[STRING_MAXLEN + 1];
  FILE *fp;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;

  fp = integer ? (FILE *)integer : stdin;
  if (!fgets(buffer, sizeof(buffer), fp))
    return NLL_ERRCODE_FORMULA_INVALID_PARAMETER;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_string(value, buffer, -1)) < 0)
    return r;

  return 0;
}

static int function_fputs(value_t value, value_t args)
{
  int r, length;
  integer_t integer;
  char *string;
  FILE *fp;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_string(args, &string, &length)) < 0)
    return r;

  fp = integer ? (FILE *)integer : stdout;
  if (!fputs(string, fp))
    return NLL_ERRCODE_FORMULA_INVALID_PARAMETER;
  fflush(fp);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, length)) < 0)
    return r;

  return 0;
}

static int function_fread(value_t value, value_t args)
{
  int r;
  area_t area;
  integer_t integer, offset, size;
  FILE *fp;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &area)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &offset)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &size)) < 0)
    return r;

  if (offset + size > area_get_size(area))
    return NLL_ERRCODE_AREA_TOO_LONG;

  fp = integer ? (FILE *)integer : stdin;
  if (fread(area_get_buffer(area) + offset, 1, size, fp) != size)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, size)) < 0)
    return r;

  return 0;
}

static int function_fwrite(value_t value, value_t args)
{
  int r;
  area_t area;
  integer_t integer, offset, size;
  FILE *fp;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &integer)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_area(args, &area)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &offset)) < 0)
    return r;
  args = args->next;

  if (!args)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_get_integer(args, &size)) < 0)
    return r;

  if (offset + size > area_get_size(area))
    return NLL_ERRCODE_AREA_TOO_LONG;

  fp = integer ? (FILE *)integer : stdout;
  if (fwrite(area_get_buffer(area) + offset, 1, size, fp) != size)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  fflush(fp);

  if (!value)
    return NLL_ERRCODE_FORMULA_LESS_PARAMETER;
  if ((r = value_set_integer(value, size)) < 0)
    return r;

  return 0;
}
