#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#ifdef USE_CURSES
#include <curses.h>
#else
#ifndef WIN32
#include <termios.h>
#else
#include <windows.h>
#include <conio.h>
#endif
#endif

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "term.h"

static int savenum = 0;

#ifdef USE_CURSES
int nll_term_start(void)
{
#if 0
  initscr();
  noecho();
  cbreak();
#endif
  return 0;
}

int nll_term_restore(void)
{
#if 0
  refresh();
  nocbreak();
  echo();
  endwin();
#endif
  return 0;
}
#else
#ifndef WIN32
static struct termios save;
static int savefl;

int nll_term_start(void)
{
  struct termios sets;

  tcgetattr(STDIN_FILENO, &sets);

  save = sets;
  sets.c_lflag &= ~(ECHO | ICANON);
  sets.c_cc[VTIME] = 0;
  sets.c_cc[VMIN]  = 1;
  tcsetattr(STDIN_FILENO, TCSANOW, &sets);

  savefl = fcntl(STDIN_FILENO, F_GETFL, 0);
  fcntl(STDIN_FILENO, F_SETFL, savefl | O_NONBLOCK);

  savenum = 0;

  return 0;
}

int nll_term_restore(void)
{
  fcntl(STDIN_FILENO, F_SETFL, savefl);
  tcsetattr(STDIN_FILENO, TCSANOW, &save);
  return 0;
}
#else
int nll_term_start(void)
{
  HANDLE handle = (HANDLE)_get_osfhandle(STDOUT_FILENO);
  DWORD mode = 0;
  if (!GetConsoleMode(handle, &mode))
    return -1;
#ifndef ENABLE_VIRTUAL_TERMINAL_PROCESSING
#define ENABLE_VIRTUAL_TERMINAL_PROCESSING 0x0004
#endif
  if (!SetConsoleMode(handle, mode | ENABLE_VIRTUAL_TERMINAL_PROCESSING))
    return -1;
  return 0;
}

int nll_term_restore(void)
{
  return 0;
}
#endif
#endif

int nll_term_key(void)
{
  int r, i, escape = 0, k;
#define KEY_SAVENUM 8
  static int key[KEY_SAVENUM];
  unsigned char c;

  for (i = 0; i < 64; i++) {
#ifndef WIN32
    r = read(STDIN_FILENO, &c, 1);
#else
    if (kbhit()) {
      c = getch();
      r = 1;
    } else {
      r = 0;
    }
#endif
    if (r <= 0)
      break;
    k = c;

    if (escape) {
      switch (escape) {
      case 0x1b:
	escape = 0;
	/* fall through */
      case 0x4f:
      case 0x5b:
      case 0x5b35:
      case 0x5b36:
	escape = (escape << 8) | k;
	switch (escape) {
	case NLL_TERM_KEY_END:
	case NLL_TERM_KEY_UP:
	case NLL_TERM_KEY_DOWN:
	case NLL_TERM_KEY_RIGHT:
	case NLL_TERM_KEY_LEFT:
	case NLL_TERM_KEY_PAGEUP:
	case NLL_TERM_KEY_PAGEDOWN:
	  k = escape;
	  escape = 0;
	  break;
	default:
	  continue;
	}
	break;

      default:
	escape = 0;
	continue;
      }
    } else if (k == 0x1b) {
      escape = k;
      continue;
    }

    if ((!savenum || (k != key[savenum - 1])) && (savenum < KEY_SAVENUM))
      key[savenum++] = k;
  }

  if (!savenum)
    return 0;

  k = key[0];
  for (i = 0; i < savenum - 1; i++)
    key[i] = key[i + 1];
  savenum--;

  return k;
}
